import logging
import typing

import boto3

LOGGER = logging.getLogger(__name__)


class ParameterStore:

    def __init__(self,
                 profile: typing.Optional[str] = None,
                 region: typing.Optional[str] = None):
        self._session = boto3.Session(profile_name=profile, region_name=region)
        self._client = self._session.client('ssm')

    def fetch_variables(self, variables: list) -> typing.Dict[str, str]:
        response = self._client.get_parameters(
            Names=list(variables), WithDecryption=True)
        return {param['Name']: param['Value']
                for param in response['Parameters']}



import argparse
import pathlib
import unittest

from ssm_ps_template import config


class TestCase1(unittest.TestCase):

    def setUp(self) -> None:
        self.expectation = config.Configuration(
            templates=[
                config.Template(
                    source=pathlib.Path('tests/templates/case1a.tmpl'),
                    destination=pathlib.Path('build/case1a.out')),
                config.Template(
                    source=pathlib.Path('tests/templates/case1b.tmpl'),
                    destination=pathlib.Path('build/case1b.out')),
            ],
            profile=None,
            region=None,
            verbose=False)

    def test_load_json_file(self):
        self.assertEqual(
            config.configuration_file('tests/config/case1.json'),
            self.expectation)

    def test_load_toml_file(self):
        self.assertEqual(
            config.configuration_file('tests/config/case1.toml'),
            self.expectation)

    def test_load_yaml_file(self):
        self.assertEqual(
            config.configuration_file('tests/config/case1.yaml'),
            self.expectation)


class TestCase2(unittest.TestCase):

    def test_invalid_file_type(self):
        with self.assertRaises(argparse.ArgumentTypeError):
            config.configuration_file('tests/config/case2d.ini')

    def test_missing_configuration(self):
        with self.assertRaises(argparse.ArgumentTypeError):
            config.configuration_file('tests/config/case.yaml')

    def test_missing_template_destination(self):
        with self.assertRaises(argparse.ArgumentTypeError):
            config.configuration_file('tests/config/case2a.yaml')

    def test_missing_templates(self):
        with self.assertRaises(argparse.ArgumentTypeError):
            config.configuration_file('tests/config/case2b.yaml')

    def test_missing_template_source(self):
        with self.assertRaises(argparse.ArgumentTypeError):
            config.configuration_file('tests/config/case2c.yaml')

    def test_missing_template_source_file(self):
        with self.assertRaises(argparse.ArgumentTypeError):
            config.configuration_file('tests/config/case2d.yaml')



import argparse
import dataclasses
import json
import pathlib
import typing

import toml
import yaml


@dataclasses.dataclass
class Template:
    source: pathlib.Path
    destination: pathlib.Path


@dataclasses.dataclass
class Configuration:
    templates: list[Template]
    profile: typing.Optional[str]
    region: typing.Optional[str]
    verbose: bool


def _load_configuration(value: dict) -> Configuration:
    try:
        return Configuration(
            templates=[_entry_to_template(**template)
                       for template in value['templates']],
            profile=value.get('profile'),
            region=value.get('region'),
            verbose=value.get('verbose', False))
    except KeyError as error:
        raise argparse.ArgumentTypeError(
            f'Failed to load configuration due to invalid key: {error}')


def _entry_to_template(**kwargs) -> Template:
    source = pathlib.Path(kwargs['source'])
    if not source.exists():
        raise argparse.ArgumentTypeError(
            f'Specified template {source} does not exist')
    return Template(
        source=source,
        destination=pathlib.Path(kwargs['destination']))


def configuration_file(value: str) -> Configuration:
    """Load the configuration from the specified path"""
    path = pathlib.Path(value)
    if not path.exists():
        raise argparse.ArgumentTypeError(f'{value} does not exist')

    if str(path).endswith('.json'):
        with path.open('r') as handle:
            return _load_configuration(json.load(handle))
    if str(path).endswith('.toml'):
        with path.open('r') as handle:
            return _load_configuration(toml.load(handle))
    if str(path).endswith('.yaml') or str(path).endswith('.yml'):
        with path.open('r') as handle:
            return _load_configuration(yaml.safe_load(handle))

    raise argparse.ArgumentTypeError(f'{value} is not an understood file type')



import argparse
from importlib import metadata
import logging
import time

from ssm_ps_template import config, render, ssm

LOGGER = logging.getLogger(__name__)


def parse_cli_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description='Templating for SSM Parameter Store')
    parser.add_argument('--aws-profile', action='store', help='AWS Profile')
    parser.add_argument('--aws-region', action='store', help='AWS Region')
    parser.add_argument('--verbose', action='store_true')
    parser.add_argument('config', type=config.configuration_file, nargs=1)
    return parser.parse_args()


def main():
    args = parse_cli_arguments()
    verbose = args.config[0].verbose or args.verbose
    logging.basicConfig(level=logging.DEBUG if verbose else logging.INFO)
    LOGGER.info('ssm-ps-template v%s', metadata.version('ssm-ps-template'))
    parameter_store = ssm.ParameterStore(
        profile=args.aws_profile or args.config[0].profile,
        region=args.aws_region or args.config[0].region)
    start_time = time.time()
    for template in args.config[0].templates:
        renderer = render.Renderer(
            source=template.source)
        variables = renderer.discover_variables()
        LOGGER.debug('Variables: %r', variables)
        values = parameter_store.fetch_variables(variables)
        with template.destination.open('w') as handle:
            handle.write(renderer.render(values))
        LOGGER.info('Rendered %s in %0.2f seconds',
                    template.destination, time.time() - start_time)



import pathlib
import unittest

from ssm_ps_template import render


class DiscoverVariablesTestCase(unittest.TestCase):

    def test_case1a(self):
        renderer = render.Renderer(
            pathlib.Path('tests/templates/case1a.tmpl'))
        variables = renderer.discover_variables()
        self.assertListEqual(sorted(variables),
                             ['foo/bar/baz', 'ssm_variable'])


class RenderingTestCase(unittest.TestCase):

    def test_case1a(self):
        renderer = render.Renderer(
            pathlib.Path('tests/templates/case1a.tmpl'))
        renderer.discover_variables()
        output = renderer.render({
            'ssm_variable': 'Variable',
            'foo/bar/baz': 'Corgie!'})
        with open('tests/expectations/case1a.out', 'r') as handle:
            expectation = handle.read()
        self.assertEqual(output.strip(), expectation.strip())

    def test_case1b(self):
        renderer = render.Renderer(
            pathlib.Path('tests/templates/case1b.tmpl'))
        renderer.discover_variables()
        output = renderer.render({
            '/foo/bar/baz': 'postgresql://localhost:5432/postgres',
            '/qux/corgie': 'redis://localhost:6379/0'})
        with open('tests/expectations/case1b.out', 'r') as handle:
            expectation = handle.read()
        self.assertEqual(output.strip(), expectation.strip())



import enum
import io
import logging
import pathlib
import typing

import jinja2
from jinja2 import sandbox

LOGGER = logging.getLogger(__name__)


class DiscoveryMode(enum.Enum):
    SIMPLE = 1
    COMPOUND = 2


class Renderer:

    def __init__(self, source: pathlib.Path):
        with source.open('r') as handle:
            self._source = handle.read()
        self._buffer = io.StringIO()
        self._variables = set({})

    def discover_variables(self) -> typing.List[str]:
        """Discover the variables that to look up in SSM Parameter Store

        Supports "fake" compound variables for SSM namespacing like:

            `{{ foo/bar/baz }}`

        """
        environment = sandbox.ImmutableSandboxedEnvironment()
        body = environment.preprocess(self._source)
        tokens = list(environment.lex(body))
        mode, variable = DiscoveryMode.SIMPLE, []
        for offset, token in enumerate(tokens):
            if mode == DiscoveryMode.SIMPLE \
                    and token[1] == 'operator' \
                    and token[2] == '/':
                LOGGER.debug('Starting COMPOUND mode')
                mode = DiscoveryMode.COMPOUND
                variable = [token[2]]
            elif mode == DiscoveryMode.COMPOUND \
                    and token[1] == 'operator' \
                    and token[2] == '/' or token[2] == '-':
                LOGGER.debug('Ignoring / operator in COMPOUND mode')
                variable.append(token[2])
            elif mode == DiscoveryMode.SIMPLE \
                    and token[1] == 'name' \
                    and self._next_is_compound_operator(offset, tokens):
                LOGGER.debug('Starting COMPOUND mode')
                mode = DiscoveryMode.COMPOUND
                variable = [token[2]]
            elif mode == DiscoveryMode.COMPOUND \
                    and token[1] == 'name' \
                    and not self._next_is_compound_operator(offset, tokens):
                LOGGER.debug('Finishing COMPOUND mode')
                variable.append(token[2])
                self._variables.add(''.join(variable))
                mode = DiscoveryMode.SIMPLE
            elif mode == DiscoveryMode.COMPOUND and token[1] == 'name':
                LOGGER.debug('Appending token in COMPOUND: %s', token[2])
                variable.append(token[2])
            elif mode == DiscoveryMode.SIMPLE and token[1] == 'name':
                LOGGER.debug('Appending variable in SIMPLE: %s', token[2])
                self._variables.add(token[2])
        return list(self._variables)

    def render(self, values: typing.Dict) -> str:
        """Render the template to the internal buffer"""
        variables = {}
        for key, value in values.items():
            variables[self._sanitize_variable(key)] = value
        return jinja2.Template(
            self._patched_source(), autoescape=True).render(**variables)

    @staticmethod
    def _next_is_compound_operator(
            offset: int,
            tokens: typing.List[typing.Tuple[int, str, str]]) -> bool:
        """Check to see if the next token is a compound operator"""
        return tokens[offset + 1][1] == 'operator' \
            and (tokens[offset + 1][2] == '/' or tokens[offset + 1][2] == '-')

    def _patched_source(self) -> str:
        """Replace `foo/bar/baz` variables with `foo__bar__baz`"""
        source = str(self._source)
        for var in list(self._variables):
            if var != self._sanitize_variable(var):
                source = source.replace(var, self._sanitize_variable(var))
        return source

    @staticmethod
    def _sanitize_variable(value: str) -> str:
        return value.replace('/', '___').replace('-', '_')
